# will be used when other operations are implemented
operations = {
    'intersection': 'ST_Intersection',
    'union': 'ST_Union',
    'difference': 'ST_Difference',
    'symmetric_difference': 'ST_SymDifference',
    'intersects': 'ST_Intersects',
}
